@@ -1,13 +1,14 @@
 from contextlib import contextmanager
 import hashlib
 import importlib
-from pathlib import Path
+import itertools
 import os
 import re
+import tempfile
 
-from django.apps import apps
 from django.conf import settings
 from django.db import connections
+from django.db.migrations.loader import MigrationLoader
 from django.test.utils import setup_databases
 
 
@@ -23,7 +24,7 @@ def _get_db_backend(connection):
 def create_test_db(*, verbosity=1):
     for connection in get_unique_connections():
         connection.creation._create_test_db(
-            verbosity=True, autoclobber=True, keepdb=False
+            verbosity=verbosity, autoclobber=True, keepdb=False
         )
 
 
@@ -38,13 +39,17 @@ def get_unique_connections():
 def setup_test_db(*, verbosity=1):
     # Based on https://github.com/django/django/blob/d62563cbb194c420f242bfced52b37d6638e67c6/django/test/runner.py#L1051-L1054  # noqa: E501
     unique_connections = get_unique_connections()
-    aliases = []
-    database_names = {}
-    for connection in unique_connections:
-        aliases.append(connection.alias)
-        database_names[connection.alias] = connection.settings_dict["NAME"]
-
-    setup_databases(verbosity=True, interactive=False, aliases=aliases)
+    database_names = {
+        connection.alias: connection.settings_dict["NAME"]
+        for connection in unique_connections
+    }
+    aliases = list(database_names.keys())
+    setup_databases(
+        aliases=aliases,
+        verbosity=verbosity,
+        interactive=False,
+        keepdb=True,
+    )
 
     # Django setup_databases change original settings and don't care about it
     # because it run the setup only one time and other parts of testing understand that.
@@ -67,7 +72,9 @@ def clone_test_db(connection, parallel, is_pytest=False, *, verbosity=1):
             # e.g. test_db_1, test_db_2, ...
             suffix = f"{index + 1}"
 
-        connection.creation.clone_test_db(suffix=suffix, verbosity=True, keepdb=False)
+        connection.creation.clone_test_db(
+            suffix=suffix, verbosity=verbosity, keepdb=False
+        )
 
         if connection.vendor == "sqlite":
             settings_dict = connection.creation.get_test_db_clone_settings(suffix)
@@ -138,8 +145,8 @@ def _transform_sqlite_db_name(db_name, *, suffix="", dotbug=False):
     return db_name
 
 
-def load(connection, input_file):
-    with open(input_file, "r") as f:
+def load(connection, input_file, storage):
+    with storage.open(input_file, "r") as f:
         sql = f.read()
     with connection.cursor() as cursor:
         if connection.vendor == "sqlite":
@@ -150,21 +157,55 @@ def load(connection, input_file):
             cursor.execute(sql)
 
 
-def dump(connection, output_file):
+def dump(connection, output_file, storage):
     backend = _get_db_backend(connection)
-    backend.dump(connection, output_file)
+
+    # Dump to a temp file because backends expect a filename instead of a file object.
+    _, tmp_filename = tempfile.mkstemp(prefix="migrateci", suffix=".sql")
+    backend.dump(connection, tmp_filename)
+
+    # Copy it to the storage.
+    with (
+        open(tmp_filename, "r") as tmp_fp,
+        storage.open(output_file, "w") as output_fp,
+    ):
+        output_fp.write(tmp_fp.read())
 
 
-def hash_files(*files):
-    files = list(files)
-    for app_config in apps.app_configs.values():
-        path = Path(app_config.path)
-        for migration_file in path.glob("migrations/*.py"):
-            files.append(migration_file)
+def hash_files(depth=0):
+    """
+    Generate checksums based on migrations graph.
 
-    checksum = hashlib.md5()
-    for _file in sorted(files):
-        with open(_file, "rb") as f:
-            checksum.update(f.read())
+    Yield many checksums first based on all migrations, after that it tries to
+    remove `depth` migrations per app.
+    """
+    loader = MigrationLoader(None, ignore_no_migrations=True)
+    nodes = loader.graph.leaf_nodes()
+    plan = loader.graph._generate_plan(nodes, at_end=True)
 
-    return checksum.hexdigest()
+    checksums = {}
+    for app_label, migration_name in plan:
+        migration = loader.get_migration(app_label, migration_name)
+        module = importlib.import_module(migration.__module__)
+        filename = module.__file__
+
+        # Calculate checksum for each migration to limit memory usage.
+        with open(filename, "rb") as f:
+            checksum = hashlib.md5(f.read()).hexdigest()
+
+        app_checksums = checksums.setdefault(app_label, [])
+        app_checksums.append(checksum)
+
+    for n in range(0, depth + 1):
+        combinations = itertools.combinations_with_replacement(checksums, n)
+        for apps_to_remove_migration in combinations:
+            copy_checksums = checksums.copy()
+            for app_label in apps_to_remove_migration:
+                # Remove last migration from selected app.
+                remaining_checksums = copy_checksums[app_label][:-1]
+                copy_checksums[app_label] = remaining_checksums
+
+            checksum = hashlib.md5()
+            for app_checksums in copy_checksums.values():
+                checksum.update("".join(app_checksums).encode())
+            yield checksum.hexdigest()
